/* *****************************************************************************
 * type/cas.c -- get the CAS type and data types out of raw identification data.
 * Copyright (C) 2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libg1m.
 * libg1m is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libg1m is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libg1m; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libg1m/internals.h>

/* ************************************************************************** */
/*  Local types                                                               */
/* ************************************************************************** */
/* Correspondance type */
struct type_corresp {
	/* identification */
	const char *datatype;

	/* libg1m MCS file type */
	unsigned int type;
};

/* ************************************************************************** */
/*  Correspondances                                                           */
/* ************************************************************************** */
/* All correspondances. Remarks:
 * - Correspondances with a NULL data type means the data type isn't to be
 *   read. */

#define TTERM {NULL, 0}
static struct type_corresp cas_groups[] = {
	{"PG", g1m_mcstype_program},

	/* not implemented yet types, described by Tom Wheeley and Tom Lynn */
	{"AA", 0}, // dynamic graph functions
	{"AD", 0}, // variable memory
	{"AL", 0}, // all
	{"AM", 0}, // alpha variable memory
	{"BU", 0}, // backup
	{"DM", 0}, // defined memory
	{"DD", 0}, // screenshot
	{"EN", 0}, // one editor file
	{"FN", 0}, // set of editor files
	{"F1", 0}, // one function memory
	{"F6", 0}, // set of function memories
	{"GA", 0}, // set of graph functions
	{"GF", 0}, // graph zoom factor (graph function?)
	{"GR", 0}, // graph range
	{"GT", 0}, // function table
	{"MA", 0}, // set of matrices
	{"PD", 0}, // polynomial equations
	{"P1", 0}, // one program
	{"PZ", 0}, // set of programs
	{"RT", 0}, // recursion table
	{"SD", 0}, // simultaneous equations
	{"SR", 0}, // linear regression data
	{"SS", 0}, // standard deviation data

	/* not implemented yet types, described in libcafix */
	{"FT", 0}, // ???
	{"GM", 0}, // G-Mem?
	{"LT", 0}, // List
	{"MT", 0}, // Matrix
	{"PC", 0}, // Picture
	{"RF", 0}, // ???
	{"RR", 0}, // ???
	{"SE", 0}, // Equation?
	{"TR", 0}, // ???
	{"VM", 0}, // Variable (A-Z, etc)
	{"WD", 0}, // window data

	/* terminating entry */
	TTERM
};

/* ************************************************************************** */
/*  Main functions                                                            */
/* ************************************************************************** */
/**
 *	g1m_castype_get:
 *	Get the libg1m MCS type from raw CAS identification data.
 *
 *	@arg	datatype	the data type string.
 *	@arg	type[out]	the libg1m MCS type.
 *	@return				the error (if any).
 */

int g1m_castype_get(const char *datatype, unsigned int *type)
{
	struct type_corresp *c = cas_groups - 1;
	while ((++c)->datatype) {
		if (c->datatype && strncmp(datatype, c->datatype, 4))
			continue;
		break;
	}
	if (!c->type) goto notfound;

	/* fill in info and return */
	if (type) *type = c->type;
	return (0);

notfound:
	log_info("Type with '%.4s' data string was not implemented or not "
		"recognized.", datatype);
	if (type) *type = 0;
	return (1);
}
